#include <stdio.h>
#include <mpi.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <memory.h>

#define RQST 1
#define WORK 2
#define SAVE 3

int main(int argc, char ** argv) {

//The only argument taken here is a positive integer N, so that a AN.bin and BN.bin can be loaded

	int my_rank;
	int world_size;
        MPI_Status status;

        int N = atoi(argv[1]); //size of matrix is N by N
	int n = N*N; //number of entries is n

	int opencomm = 1;
	int tally = 2; // Used to terminate the last request for work

	int Row_Entry = 0; // Used to count number of rows computed, each node has a version of this, for different reasons
	int Col_Entry = 0; // Used to count number of columns computed (slave nodes only)
	int Col_Count = 0; // Used to count number of columns sent to slave nodes from node 0

	// Flag[0] = -1 request a row of A, Flag[0] = 1 request a row of B
	int Flag[2]; 

	// Place holder for Matrix data which will be sent from node 0 to nodes > 1
	int *AData;
	int *BData;
	int *CData;

	AData = (int *)malloc((N+1)*sizeof(int));
	BData = (int *)malloc((N+1)*sizeof(int));
	CData = (int *)malloc((N+1)*sizeof(int));

        FILE *fptA;
        FILE *fptB;
        FILE *fptBT;
        FILE *fptC;

        char fnameA[80];
        strcpy(fnameA, "A");
        strcat(fnameA, argv[1]);
        strcat(fnameA, ".bin");

        char fnameB[80];
        strcpy(fnameB, "B");
        strcat(fnameB, argv[1]);
        strcat(fnameB, ".bin");

        char fnameC[80];
        strcpy(fnameC, "C");
        strcat(fnameC, argv[1]);
        strcat(fnameC, ".bin");

        char fnameBT[80];
        strcpy(fnameBT, "BT");
        strcat(fnameBT, argv[1]);
        strcat(fnameBT, ".bin");

	MPI_Init(&argc, &argv); /* note that argc and argv are passed by address */

        MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
        MPI_Comm_size(MPI_COMM_WORLD, &world_size);

	if (my_rank == 0)
	{
		printf("Node 0: Initializing. \n");
//              Here we performing the transpose of B before we get started on anything else...
                printf("We are performing the tranpose of B\n");

                fptB = fopen(fnameB, "rb");
                fptBT = fopen(fnameBT, "wb");

                int cntr = 0;
	
		int *BTData;
                int jj;
		int kk;

		BTData = (int *)malloc((N+1)*sizeof(int));

                for(jj=0; jj<=N-1;jj++)
		{
                        for(kk=0; kk<=N-1;kk++)
			{
                                fseek(fptB, (kk*N+jj)*4, 0);
                                fread(&BTData[kk], 4, 1, fptB);
//                              printf("Entry %d of the matrix is %d\n", cntr, BTData[kk]);
                        }
                        fseek(fptBT, cntr*4*N, SEEK_SET);
                        fwrite(BTData, 4, N, fptBT);
                        cntr++;
                }
                fclose(fptB);
                fclose(fptBT);

                printf("Transpose of B complete\n");

                fptA = fopen(fnameA, "rb");
                fptBT = fopen(fnameBT, "rb");

		while(opencomm)
		{
			// Receive request from a node > 1 to do work
//			printf("Node 0: Row_Entry = %d, Col_Count = %d, Sum = %d, looking for %d.  \n", Row_Entry, Col_Count, Row_Entry + Col_Count, N+n); 
			MPI_Recv(&Flag, 2, MPI_INT, MPI_ANY_SOURCE, RQST, MPI_COMM_WORLD, &status);
			
			if (Flag[0] == -1) // this means to send a row of A 
			{
				if (Row_Entry < N) // there is still rows of A to send off
				{
					AData[0] = Row_Entry; // rest of array should be row of A
					fread(&AData[1], 4, N, fptA);
					
//					printf("Node 0: First entry in row %d of A is %d. \n", AData[0], AData[1]);

					MPI_Send(AData, N+1, MPI_INT, status.MPI_SOURCE, WORK, MPI_COMM_WORLD);
//					printf("Node 0: Sending row %d of %d from Matrix A to Node %d. \n", Row_Entry, N, status.MPI_SOURCE);
					Row_Entry++; // increment Row_Entry to keep track of how many of the N rows of A have been sent
				}
				else // no more rows of A to send
				{
					AData[0] = -2; // Set row entry to -1 to stop other nodes from requesting
					tally++; // increment tally to detemine how many nodes have been told to stop
					MPI_Send(AData, N+1, MPI_INT, status.MPI_SOURCE, WORK, MPI_COMM_WORLD);
//					printf("Node 0: Telling Node %d that there is no more work. \n", status.MPI_SOURCE);
				}

			}
			else if (Flag[0] == 1) // this means to send a row of B
			{
				BData[0] = Flag[1]; // make sure that the correct column of B is sent
	                        fseek(fptBT,  BData[0]*N*sizeof(int), SEEK_SET);
        	                fread(&BData[1], 4, N, fptBT);

//				printf("Node 0: First entry in column %d of B is %d. \n", BData[0], BData[1]);

				MPI_Send(BData, N+1, MPI_INT, status.MPI_SOURCE, WORK, MPI_COMM_WORLD);
//				printf("Node 0: Sending row %d of Matrix B to Node %d. \n", BData[0], status.MPI_SOURCE);
				Col_Count++; // increment number of columns sent to slave nodes
			}

			// There should be N rows of A sent, and N*N = n columns of C sent, since each column of C must be sent N times
			// Tally gets to world_size  since it starts at 2 and increments each time a node is told to stop
			if (Row_Entry + Col_Count + tally  == N + n + world_size) // then there is no more work to do, so quit!
			{
				opencomm = 0;
				printf("Node 0: All work has been dished out, time for dinner.\n");
			}
		}

                fclose(fptBT);
                fclose(fptA);
                remove(fnameBT);
	}
	else if (my_rank == 1)
	{
                fptC = fopen(fnameC, "wb");

		printf("Node 1: Initializing. \n");
		while(Row_Entry < N) // keep receiving rows from nodes > 1 until all N rows have been accounted for
		{
			MPI_Recv(CData, N+1, MPI_INT, MPI_ANY_SOURCE, SAVE, MPI_COMM_WORLD, &status);
//			printf("Node 1: Received row %d of Matrix C from Node %d. \n", CData[0], status.MPI_SOURCE);
		
                        fseek(fptC, CData[0]*N*sizeof(int), 0);
                        fwrite(&CData[1], 4, N, fptC);

			Row_Entry++;
		}			
		printf("Node 1: All work has been done. \n");
		fclose(fptC);
	}
	else
	{
		int jj = 0;
		int kk = 0;
		printf("Node %d: Initializing. \n", my_rank);
		Flag[0] = -1; // initally set to -1 to grab a row of A
		while(opencomm)
		{
			MPI_Send(&Flag, 2, MPI_INT, 0, RQST, MPI_COMM_WORLD); // request a row of A
//			printf("Node %d: Sending request to Node 0 for a row of Matrix A. \n", my_rank);	
	
			MPI_Recv(AData, N+1, MPI_INT, 0, WORK, MPI_COMM_WORLD, &status); // receive a row of A
//			printf("Node %d: Received row %d of Matrix A from Node 0.\n", my_rank, AData[0]);
			// AData[0] has the row number of the row A sent			

//			printf("Node %d: First entry in row %d of A is %d. \n", my_rank, AData[0], AData[1]);

			if (AData[0] == -2) // if row number of A is -2, then stop working!
			{
				printf("Node %d: Received message to stop work from Node 0. \n", my_rank);
				opencomm = 0;	
			}
			else // otherwise, go through all the columns of B
			{
				Row_Entry++;
				CData[0] = AData[0]; // Make sure that when we send row of C to node 1 it knows where to store it 
				for (Col_Entry = 0; Col_Entry < N; Col_Entry++)
				{
					Flag[0] = 1; // set to 1 to reqest a column of B
					Flag[1] = Col_Entry; // column number to request

					MPI_Send(&Flag, 2, MPI_INT, 0, RQST, MPI_COMM_WORLD);
//					printf("Node %d: Sending request to Node 0 for column %d of Matrix B. \n", my_rank, Col_Entry);	

					MPI_Recv(BData, N+1, MPI_INT, 0, WORK, MPI_COMM_WORLD, &status);
//					printf("Node %d: Received column %d of Matrix B from Node 0. \n", my_rank, BData[0]);
//					printf("Node %d: First entry in column %d of B is %d. \n", my_rank, BData[0], BData[1]);
	
					CData[Col_Entry+1] = 0;
					for(jj=1; jj<=N; jj++)
					{
						CData[Col_Entry+1] = CData[Col_Entry+1] + AData[jj]*BData[jj];
					}
//					printf("Node %d: Computed entry %d of row %d with a value of %d.\n", my_rank, Col_Entry, CData[0], CData[Col_Entry+1]);
					// Do a for loop to compute the entry Col_Entry in the row of C to send
				}

				// Now that all columns of B have been sent to this node, we can send row of C to Node 1
				MPI_Send(CData, N+1, MPI_INT, 1, SAVE, MPI_COMM_WORLD);
//				printf("Node %d: Sending row %d of Matrix C to Node 1. \n", my_rank, CData[0]);
				Flag[0] = -1;
			}
		}
		printf("Node %d: Computed %d rows of Matrix C. \n", my_rank, Row_Entry);
	}
    	MPI_Finalize();
}
